﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Runtime.InteropServices;
using System.Security.Cryptography;
using System.Text;

namespace Touchlife.Api
{
    public abstract class SecurityUtils
    {
        public static string GenerateSign(IDictionary<string, object> parameters, string secret, string privateKey)
        {
            string stringToSigned = ParamsTostring(parameters, secret);
            string signType = (string)parameters["signType"];

            if (Constants.SIGN_TYPE_MD5.Equals(signType)) 
            {
                return SignWithMD5(stringToSigned);
            } 
            else if (Constants.SIGN_TYPE_RSA.Equals(signType))
            {
                return SignWithRSA(stringToSigned, privateKey);
            }

            return null;
        }

        public static bool VerifySign(IDictionary<string, object> parameters, string secret, string publicKey)
        {
            string stringToSigned = ParamsTostring(parameters, secret);
            string signType = (string)parameters["signType"];
            string sign = (string)parameters["sign"];

            bool res = false;
            if (Constants.SIGN_TYPE_MD5.Equals(signType)) 
            {
                res = string.Equals(sign, SignWithMD5(stringToSigned));
            }
            else if (Constants.SIGN_TYPE_RSA.Equals(signType))
            {
                res = VerifyWithRSA(stringToSigned, sign, publicKey);
            }

            if (!res)
            {
                throw new ApiException(Constants.ERROR_CODE_SIGNATURE_INVALID, "server-error:Invalid signature");
            }

            return res;
        }

        private static string ParamsTostring(IDictionary<string, object> parameters, string secret)
        {
        	string[] keys = new string[parameters.Keys.Count];
        	parameters.Keys.CopyTo(keys, 0);
        	Array.Sort(keys);
        	StringBuilder sb = new StringBuilder();
        	foreach(string key in keys)
        	{
        		if(DefaultTouchlifeClient.SIGN != key && DefaultTouchlifeClient.SIGN_TYPE != key)
        		{
        			var strValue = parameters[key];
					if(null != strValue && string.Empty != strValue)
        			{
        				sb.Append(string.Format("{0}={1}&",key, strValue));
        			}
        		}
        	}
        	sb.Append("appsecret=" + secret);
        	return sb.ToString();
        }

        private static string SignWithMD5(string data)
        {
            MD5 md5 = MD5.Create();
            byte[] bytes = md5.ComputeHash(Encoding.UTF8.GetBytes(data));

            StringBuilder result = new StringBuilder();
            for (int i = 0; i < bytes.Length; i++)
            {
                result.Append(bytes[i].ToString("X2"));
            }

            return result.ToString();
        }
        
        public static string SignWithRSA(string data, string privateKeyPem)
        {
            RSACryptoServiceProvider rsaCsp = LoadCertificateFile(privateKeyPem);
            byte[] dataBytes = Encoding.GetEncoding(Constants.CHARSET_UTF8).GetBytes(data);
            byte[] signatureBytes = rsaCsp.SignData(dataBytes, "SHA1");
            return Convert.ToBase64String(signatureBytes);
        }
        
		public static bool VerifyWithRSA(string content, string sign, string publicKeyPem)
		{
            try
            {
                string sPublicKeyPEM = File.ReadAllText(publicKeyPem);
                RSACryptoServiceProvider rsa = new RSACryptoServiceProvider();
                rsa.PersistKeyInCsp = false;
                RSACryptoServiceProviderExtension.LoadPublicKeyPEM(rsa, sPublicKeyPEM);
                SHA1CryptoServiceProvider sha1 = new SHA1CryptoServiceProvider();
                bool bVerifyResultOriginal = rsa.VerifyData(Encoding.GetEncoding(Constants.CHARSET_UTF8).GetBytes(content), sha1, Convert.FromBase64String(sign));
                return bVerifyResultOriginal;
            }
            catch(Exception ex)
            {
            	throw new ApiException(ex.Message, ex);
            }
		}
        
        private static byte[] GetPem(string type, byte[] data)
        {
            string pem = Encoding.UTF8.GetString(data);
            string header = String.Format("-----BEGIN {0}-----\\n", type);
            string footer = String.Format("-----END {0}-----", type);
            int start = pem.IndexOf(header) + header.Length;
            int end = pem.IndexOf(footer, start);
            string base64 = pem.Substring(start, (end - start));
            return Convert.FromBase64String(base64);
        }
        
        private static RSACryptoServiceProvider LoadCertificateFile(string filename)
        {
            using (System.IO.FileStream fs = System.IO.File.OpenRead(filename))
            {
                byte[] data = new byte[fs.Length];
                byte[] res = null;
                fs.Read(data, 0, data.Length);
                if (data[0] != 0x30)
                {
                    res = GetPem("RSA PRIVATE KEY", data);
                }
                try
                {
                    RSACryptoServiceProvider rsa = DecodeRSAPrivateKey(res);
                    return rsa;
                }
                catch
                {
                }
                return null;
            }
        }
        
 		private static RSACryptoServiceProvider DecodeRSAPrivateKey(byte[] privkey)
        {
            byte[] MODULUS, E, D, P, Q, DP, DQ, IQ;

            // --------- Set up stream to decode the asn.1 encoded RSA private key ------
            MemoryStream mem = new MemoryStream(privkey);
            BinaryReader binr = new BinaryReader(mem);  //wrap Memory Stream with BinaryReader for easy reading
            byte bt = 0;
            ushort twobytes = 0;
            int elems = 0;
            try
            {
                twobytes = binr.ReadUInt16();
                if (twobytes == 0x8130) //data read as little endian order (actual data order for Sequence is 30 81)
                    binr.ReadByte();    //advance 1 byte
                else if (twobytes == 0x8230)
                    binr.ReadInt16();    //advance 2 bytes
                else
                    return null;

                twobytes = binr.ReadUInt16();
                if (twobytes != 0x0102) //version number
                    return null;
                bt = binr.ReadByte();
                if (bt != 0x00)
                    return null;


                //------ all private key components are Integer sequences ----
                elems = GetIntegerSize(binr);
                MODULUS = binr.ReadBytes(elems);

                elems = GetIntegerSize(binr);
                E = binr.ReadBytes(elems);

                elems = GetIntegerSize(binr);
                D = binr.ReadBytes(elems);

                elems = GetIntegerSize(binr);
                P = binr.ReadBytes(elems);

                elems = GetIntegerSize(binr);
                Q = binr.ReadBytes(elems);

                elems = GetIntegerSize(binr);
                DP = binr.ReadBytes(elems);

                elems = GetIntegerSize(binr);
                DQ = binr.ReadBytes(elems);

                elems = GetIntegerSize(binr);
                IQ = binr.ReadBytes(elems);
                

                // ------- create RSACryptoServiceProvider instance and initialize with public key -----
                CspParameters CspParameters = new CspParameters();
                CspParameters.Flags = CspProviderFlags.UseMachineKeyStore;
                RSACryptoServiceProvider RSA = new RSACryptoServiceProvider(1024, CspParameters);
                RSAParameters RSAparams = new RSAParameters();
                RSAparams.Modulus = MODULUS;
                RSAparams.Exponent = E;
                RSAparams.D = D;
                RSAparams.P = P;
                RSAparams.Q = Q;
                RSAparams.DP = DP;
                RSAparams.DQ = DQ;
                RSAparams.InverseQ = IQ;
                RSA.ImportParameters(RSAparams);
                return RSA;
            }
            catch
            {
                return null;
            }
            finally
            {
                binr.Close();
            }
        }
        
		private static int GetIntegerSize(BinaryReader binr)
		{
            byte bt = 0;
            byte lowbyte = 0x00;
            byte highbyte = 0x00;
            int count = 0;
            bt = binr.ReadByte();
            if (bt != 0x02)        //expect integer
                return 0;
            bt = binr.ReadByte();

            if (bt == 0x81)
            {
                count = binr.ReadByte();    // data size in next byte
            }
            else
            {
                if (bt == 0x82)
                {
                    highbyte = binr.ReadByte();    // data size in next 2 bytes
                    lowbyte = binr.ReadByte();
                    byte[] modint = { lowbyte, highbyte, 0x00, 0x00 };
                    count = BitConverter.ToInt32(modint, 0);
                }
                else
                {
                    count = bt;        // we already have the data size
                }
            }

            while (binr.ReadByte() == 0x00)
            {    //remove high order zeros in data
                count -= 1;
            }
            binr.BaseStream.Seek(-1, SeekOrigin.Current);        //last ReadByte wasn't a removed zero, so back up a byte
            return count;
		}

    }
}
